Package: DPQ
Title: Density, Probability, Quantile ('DPQ') Computations
Version: 0.5-2
Date: 2022-06-08
Authors@R: c(person("Martin","Maechler", role=c("aut","cre"),
                    email="maechler@stat.math.ethz.ch", comment = c(ORCID = "0000-0002-8685-9910"))
  , person("Morten",   "Welinder",    role = "ctb", comment = "pgamma C code, see PR#7307, Jan. 2005; further pdhyper()")
  , person("Wolfgang", "Viechtbauer", role = "ctb", comment = "dtWV(), 2002")
  , person("Ross",     "Ihaka",       role = "ctb", comment = "src/qchisq_appr.c")
  , person("Marius",   "Hofert",      role = "ctb", comment = "lsum(), lssum()")
  , person("R-core", email = "R-core@R-project.org", role = "ctb", comment = "src/{dpq.h, algdiv.c, pnchisq.c, bd0.c}")
  , person("R Foundation", role = "cph", comment = "src/qchisq-appr.c")
  	     )
Description: Computations for approximations and alternatives for the 'DPQ'
  (Density (pdf), Probability (cdf) and Quantile) functions for probability
  distributions in R.
  Primary focus is on (central and non-central) beta, gamma and related
  distributions such as the chi-squared, F, and t.
  --
  This is for the use of researchers in these numerical approximation
  implementations, notably for my own use in order to improve standard
  R pbeta(), qgamma(), ..., etc: {'"dpq"'-functions}.
Depends: R (>= 3.6.0)
Imports: stats, graphics, methods, utils, sfsmisc (>= 1.1-13)
Suggests: Rmpfr, DPQmpfr (>= 0.3-1), gmp, MASS, mgcv, scatterplot3d,
        interp
SuggestsNote: MASS::fractions() in ex | mgcv, scatt.., interp: some
        tests/
License: GPL (>= 2)
Encoding: UTF-8
URL: https://specfun.r-forge.r-project.org/,
        https://r-forge.r-project.org/R/?group_id=611
BugReports: https://r-forge.r-project.org/tracker/?atid=2462&group_id=611
NeedsCompilation: yes
Packaged: 2022-06-10 21:27:20 UTC; maechler
Author: Martin Maechler [aut, cre] (<https://orcid.org/0000-0002-8685-9910>),
  Morten Welinder [ctb] (pgamma C code, see PR#7307, Jan. 2005; further
    pdhyper()),
  Wolfgang Viechtbauer [ctb] (dtWV(), 2002),
  Ross Ihaka [ctb] (src/qchisq_appr.c),
  Marius Hofert [ctb] (lsum(), lssum()),
  R-core [ctb] (src/{dpq.h, algdiv.c, pnchisq.c, bd0.c}),
  R Foundation [cph] (src/qchisq-appr.c)
Maintainer: Martin Maechler <maechler@stat.math.ethz.ch>
Repository: CRAN
Date/Publication: 2022-06-10 22:00:09 UTC
Built: R 4.0.5; x86_64-w64-mingw32; 2022-06-11 04:25:52 UTC; windows
Archs: x64
