# DPQ/safe/ : Saved  R  objects in order to save time in tests etc

   Size | Filename                  | used-in                    | description
------: |----------                 | -------                    | -----------
   2443 | pnt-prec-sim1.rds         |                            |
 207675 | tests_chisq-nonc-ssAp.rds |../../tests/chisq-nonc-ex.R | 4D-array of non-centr pchisq() approx
  45936 | tests_chisq-nonc-ssR.rds  |../../tests/chisq-nonc-ex.R |
  99732 | tests_chisq-nonc-ssR2.rds |../../tests/chisq-nonc-ex.R |
  41102 | tests_chisq-nonc-ssR3.rds |../../tests/chisq-nonc-ex.R |
  88986 | tests_chisq-nonc-ssR4.rds |../../tests/chisq-nonc-ex.R |
    605 | tests_ppoisErr.rda		|../../tests/ppois-ex/R      |
   2740 | tests_qbeta-d-ssR.rds	    |../../tests/qbeta-dist.R    |
